﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------
	
-------------------------------------------------------------------------------
-- local cache

local CreateFrame    = CreateFrame;
local SetButtonPulse = SetButtonPulse;				
local UIFrameFlash   = UIFrameFlash;

-------------------------------------------------------------------------------

nUI_MicroMenuButtons =
{
	[1] = { name = "CharacterMicroButton", action = "TOGGLECHARACTER0", group = "Character", },
	[2] = { name = "SpellbookMicroButton", action = "TOGGLESPELLBOOK",  group = "Spellbook", },
	[3] = { name = "TalentMicroButton",    action = "TOGGLETALENTS",    group = "Talents", },
	[4] = { name = "QuestLogMicroButton",  action = "TOGGLEQUESTLOG",   group = "Quest", },
	[5] = { name = "SocialsMicroButton",   action = "TOGGLESOCIAL",     group = "Socials", },
	[6] = { name = "LFGMicroButton",       action = "TOGGLELFGPARENT",  group = "LFG", },
	[7] = { name = "MainMenuMicroButton",  action = "TOGGLEGAMEMENU",   group = "MainMenu", },
	[8] = { name = "HelpMicroButton",      action = "TOGGLEGAMEHELP",   group = "Help", },
};

nUI.Class.ButtonBar = nUI.Class:CreatePrototype();
local prototype     = nUI.Class.ButtonBar.prototype;

-------------------------------------------------------------------------------

function prototype:init( parent, id, layout )
	
	self.parent      = parent;
	self.id          = id;
	self.rid         = 1;
	
	self:layoutFrame( layout, true );
	
end

-------------------------------------------------------------------------------

function prototype:layoutFrame( layout, is_init )

	if not is_init then self:unregisterEvents(); end
	
	self.layout      = layout;
	self.active_bar  = self;
	
	local row_anchor = nil;
	local last_btn   = nil;
	local show_grid  = "no";

	-- if we haven't created this object yet, then create its frames
	
	if is_init then

		self.Buttons     = {};
		
		self.frame            = CreateFrame( "Frame", self.layout.name, self.parent );
		self.frame.flash      = CreateFrame( "Frame", self.layout.name.."_Flash", self.frame );
		self.frame.label      = CreateFrame( "Frame", self.layout.name.."_Label", self.frame );
		self.frame.label.text = self.frame.label:CreateFontString( self.layout.name.."_LabelText", "OVERLAY" );
		
	end
	
	-- create the backdrop for the bar
	
	self.frame.class = self;

	self.frame:SetScale( 1.0 / nUI.scale );
	self.frame:SetFrameStrata( "BACKGROUND" );
	self.frame:SetFrameLevel( 3 );
	self.frame:SetBackdrop(
		{
			bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
			edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
			tile     = true, 
			tileSize = 1, 
			edgeSize = 12, 
			insets   = {left = 0, right = 0, top = 0, bottom = 0},
		}
	);

	self.frame:SetBackdropColor( 0, 0, 0, 0.35 );

	if self.layout.name == "nUI_MicroMenu" then
		self.frame:SetPoint( "TOP", self.parent, "TOP", self.layout.Origin.xOfs, self.layout.Origin.yOfs );
		self.frame:SetWidth( ((self.layout.btn_size * 18 / 25 + self.layout.btn_gap) * self.layout.cols - self.layout.btn_gap + 8) * nUI.scale );
		self.frame:SetHeight( ((self.layout.btn_size + self.layout.btn_gap) * self.layout.rows - self.layout.btn_gap + 8) * nUI.scale );
	else
		self.frame:SetPoint( self.layout.Origin.anchor, self.parent, "CENTER", self.layout.Origin.xOfs * nUI.scale, self.layout.Origin.yOfs * nUI.scale );	
		self.frame:SetWidth( ((self.layout.btn_size + self.layout.btn_gap) * self.layout.cols - self.layout.btn_gap + 8) * nUI.scale );
		self.frame:SetHeight( ((self.layout.btn_size + self.layout.btn_gap) * self.layout.rows - self.layout.btn_gap + 8) * nUI.scale );
	end

	if self.layout.parent_bar and not _G[self.layout.parent_bar]:IsShown() then
		self.frame:Hide();
	end
		
	-- create a bar flash frame
		
	self.frame.flash:SetWidth( self.frame:GetWidth() + 16 );
	self.frame.flash:SetHeight( self.frame:GetHeight() + 16 );
	self.frame.flash:SetPoint( "CENTER", self.frame, "CENTER", 0, 0 );
	self.frame.flash:SetBackdrop(
		{
			bgFile   = nil,
			edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelFlash.blp", 
			tile     = true, 
			tileSize = 1, 
			edgeSize = 8, 
			insets   = {left = 0, right = 0, top = 0, bottom = 0},
		}
	);
	self.frame.flash:Hide();

	-- create a label frame
	
	self.frame.label:SetAllPoints( self.frame );
	self.frame.label:SetFrameStrata( "HIGH" );
	self.frame.label:Hide();

	self.frame.label.text:SetPoint( "CENTER", self.frame.label, "CENTER", 0, 0 );
	self.frame.label.text:SetFont( "Fonts\\\FRIZQT__.TTF", self.frame:GetHeight() * 0.4, "OUTLINE" );
	self.frame.label.text:SetTextColor( 1, 0.8196079, 0, 1 );
	self.frame.label.text:SetJustifyH( "CENTER" );
	self.frame.label.text:SetJustifyV( "MIDDLE" );
	self.frame.label.text:SetText( self.layout.label );
	
	-- add the up and down scroll buttons if this is the action bar
	
	if self.layout.name == "nUI_MainMenuBarArtFrame" then
		ActionBarUpButton:SetScale( 1.5 * nUI.scale );
		ActionBarUpButton:ClearAllPoints();
		ActionBarUpButton:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMLEFT", 10 * nUI.scale, -15 * nUI.scale );
		ActionBarDownButton:SetScale( 1.5 * nUI.scale );
		ActionBarDownButton:ClearAllPoints();
		ActionBarDownButton:SetPoint( "BOTTOMLEFT", self.frame, "BOTTOMRIGHT", -10 * nUI.scale, -15 * nUI.scale  );
	end
		
	-- add the buttons
	
	if self.layout.show_grid == "yes" then
		show_grid = "yes";
	end
	
	if self.layout.key_ring ~= "none" then
		self.layout.cols = self.layout.cols+1;
	end
		
	for i=1, self.layout.rows, 1 do
	
		for j=1, self.layout.cols, 1 do

			local micromenu_btn;
			local btn_layout      = nUI.Class.ButtonLayout:new( nil, 0 );
			local btn_id          = self.id + self.rid - 1;
			local this_id         = (i-1) * self.layout.cols + j;
			
			btn_layout.height     = self.layout.btn_size * nUI.scale;
			btn_layout.width      = self.layout.btn_size * nUI.scale;
			btn_layout.name       = self.layout.name.."_Button"..this_id;
			btn_layout.show_grid  = show_grid;
			btn_layout.hotkeys    = self.layout.hotkeys;
			btn_layout.counts     = self.layout.counts;
			btn_layout.inherits   = self.layout.inherits;

			if self.layout.name == "nUI_MicroMenu" then
				
				micromenu_btn = nUI_MicroMenuButtons[this_id];
				
				if not micromenu_btn then break; end				
				
				btn_layout.button = micromenu_btn.name;
				local button      = _G[micromenu_btn.name];				

				btn_layout.width   = btn_layout.height * 18 / 25;
				btn_layout.rescale = 1.7;
				btn_layout.safe    = true;
				
				button:SetParent( self.frame );
				
			elseif self.layout.name == "nUI_Bar_Bags" then				
				if self.layout.key_ring ~= "left" then
					if this_id == 5 then 
						btn_layout.button = "MainMenuBarBackpackButton";
					elseif this_id < 6 then 
						btn_layout.button = "CharacterBag"..(4-this_id).."Slot"; 
					else 
						btn_layout.button = "KeyRingButton"; 
						btn_layout.safe   = true;
						btn_layout.width = btn_layout.width * 0.47368421052631578947368421052632; 
						self.frame:SetWidth( self.frame:GetWidth() + btn_layout.width + self.layout.btn_gap * nUI.scale );
						self.frame.flash:SetWidth( self.frame:GetWidth() + 16 );
					end
				else
					if this_id == 6 then 
						btn_layout.button   = "MainMenuBarBackpackButton";
					elseif this_id ~= 1 then 
						btn_layout.button = "CharacterBag"..(5-this_id).."Slot"; 
					else 
						btn_layout.button = "KeyRingButton"; 
						btn_layout.safe   = true;
						btn_layout.width = btn_layout.width * 0.47368421052631578947368421052632; 
						self.frame:SetWidth( self.frame:GetWidth() + btn_layout.width + self.layout.btn_gap * nUI.scale );
						self.frame.flash:SetWidth( self.frame:GetWidth() + 16 );
					end
				end
			elseif self.layout.button_src then
				btn_layout.button = self.layout.button_src..this_id;
				btn_layout.id     = this_id;
			else

				btn_layout.button = self.layout.name.."Button"..this_id;
				btn_layout.id     = btn_id;
			end

			if self.layout.name == "nUI_MicroMenu" then
					btn_layout.btn_type = micromenu_btn.action;
			elseif self.layout.name == "nUI_Bar_Bags" then
				if self.layout.key_ring ~= "left" then
					if this_id == 5 then btn_layout.btn_type = "TOGGLEBACKPACK";
					elseif this_id < 6 then btn_layout.btn_type = "TOGGLEBAG"; 
					else btn_layout.btn_type = "TOGGLEKEYRING"; end
				else
					if this_id == 6 then btn_layout.btn_type = "TOGGLEBACKPACK";
					elseif this_id ~= 1 then btn_layout.btn_type = "TOGGLEBAG"; 
					else btn_layout.btn_type = "TOGGLEKEYRING"; end
				end
			elseif self.layout.btn_type then
				btn_layout.btn_type = self.layout.btn_type;
			else
				btn_layout.btn_type = strupper( btn_layout.button );
			end
			
			local this_btn = self.Buttons[btn_id];
			
			if not this_btn then 
				this_btn = nUI.Class.Button:new( self.frame, btn_id, btn_layout ); 
				self.Buttons[btn_id] = this_btn;				
			else 
				this_btn:layoutFrame( btn_layout ); 
			end
			
			if row_anchor == nil then
				row_anchor = this_btn;
				this_btn:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", (4 * nUI.scale), (-4 * nUI.scale) );
			elseif j == 1 then
				this_btn:SetPoint( "TOP", row_anchor.grid, "BOTTOM", 0, -self.layout.btn_gap * nUI.scale );
				row_anchor = this_btn;
			else
				this_btn:SetPoint( "LEFT", last_btn.grid, "RIGHT", self.layout.btn_gap * nUI.scale, 0 );
			end

			last_btn = this_btn;
			self.rid = self.rid + 1;

		end
	end

	self:registerEvents();
	
	if self.layout.visible == "no" then
		self.frame:Hide();
	end
	
end

-------------------------------------------------------------------------------

function prototype:registerEvents()
	
	if self.layout.name == "nUI_PossessBarFrame" then
		
		self.frame:SetScript( "OnShow",
			function()
				for i in pairs( self.Buttons ) do
					
					if i < NUM_POSSESS_SLOTS then
						self.Buttons[i]:Show();
					else
						self.Buttons[i]:Hide();
					end
				end
			end
		);
	end
	
	if self.layout.parent_bar then
	
		self.onParentShow = _G[self.layout.parent_bar]:GetScript( "OnShow" );
		self.onParentHide = _G[self.layout.parent_bar]:GetScript( "OnHide" );
			
		_G[self.layout.parent_bar]:SetScript( "OnShow",
			function()
				if self.onParentShow then self:onParentShow(); end
				self.frame:Show();
			end
		);
		
		_G[self.layout.parent_bar]:SetScript( "OnHide",
			function()
				if self.onParentHide then self:onParentHide(); end
				self.frame:Hide();
			end
		);
	end	
end	

-------------------------------------------------------------------------------

function prototype:unregisterEvents()
	
	if self.layout.name == "nUI_PossessBarFrame" then
		self.frame:SetScript( "OnShow", nil );
	end
	
	if self.layout.parent_bar then
			
		_G[self.layout.parent_bar]:SetScript( "OnShow", self.parentOnShow );
		_G[self.layout.parent_bar]:SetScript( "OnHide", self.parentOnHide );
		
	end	
end	

-------------------------------------------------------------------------------

function prototype:showGrid()

	if not self.shapeshifted then
		for i in pairs( self.Buttons ) do
			self.Buttons[i]:showGrid();
		end	
	end
end

function prototype:hideGrid()
	
	for i in pairs( self.Buttons ) do
		self.Buttons[i]:hideGrid();
	end
end

-------------------------------------------------------------------------------

function prototype:SetPoint( point, relative_to, relative_pt, xOfs, yOfs )
	
	self.frame:SetPoint( point, relative_to, relative_pt, xOfs, yOfs );
	
end

function prototype:SetAllPoints( relative_to )
	
	self.frame:SetAllPoints( relative_to );
	
end

function prototype:Show()
	
	self.frame:Show();
	
end

function prototype:Hide()
	
	self.frame:Hide();
	
end

-------------------------------------------------------------------------------

function prototype:flashBar( r, g, b, a )
	
	self.frame:SetBackdropBorderColor( r, g, b, 1 );
	self.frame.flash:SetBackdropBorderColor( r, g, b, a );
	
	UIFrameFlash( self.frame.flash, 0.5, 0.5, 8, false, 0.2, 0.2 );
end

function prototype:flashLabel()
	
	UIFrameFlash( self.frame.label, 0.5, 0.5, 10, false, 0, 9 );

end

-------------------------------------------------------------------------------

function prototype:selectBar()
	
	local action_bar = nUI.Bars["nUI_MainMenuBarArtFrame"];
	
	action_bar.active_bar:deselectBar();
	action_bar.active_bar = self;
	
	self:flashBar( 1, 0, 0, 0.75 );
	
end

function prototype:deselectBar()

	self.frame:SetBackdropBorderColor( 1, 1, 1, 1 );
	self.frame.flash:Hide();
		
end

-------------------------------------------------------------------------------

function prototype:onEnter()

	if self.layout.show_grid == "hover" then
		self:showGrid();
	end
end

-------------------------------------------------------------------------------

function prototype:onLeave()
	
	if self.layout.show_grid == "hover" then
		self:hideGrid();
	end
end
