﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------
	
-------------------------------------------------------------------------------
-- local cache

local CreateFrame            = CreateFrame;
local GetFactionInfo         = GetFactionInfo;
local GetNumFactions         = GetNumFactions;
local GetWatchedFactionInfo  = GetWatchedFactionInfo;
local SetWatchedFactionIndex = SetWatchedFactionIndex;

nUI_FactionStandings =
{
	[0] = "Unknown",	
	[1] = "Hated",	
	[2] = "Hostile",	
	[3] = "Unfriendly",	
	[4] = "Neutral",	
	[5] = "Friendly",	
	[6] = "Honored",	
	[7] = "Revered",	
	[8] = "Exalted", 
};

-------------------------------------------------------------------------------

nUI.Class.FactionBar = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local prototype      = nUI.Class.FactionBar.prototype;

-------------------------------------------------------------------------------

function prototype:layoutFrame( layout, is_init )

	if not is_init then 
		self:unregisterEvents(); 
		prototype.super.prototype.layoutFrame( self, layout );
	else
		self.event_frame = CreateFrame( "Frame", layout.name.."_EventFrame", self.parent );
		prototype.super.prototype.layoutFrame( self, layout, true );
		self.frame.grid  = CreateFrame( "Frame", "$parent_Grid", self.frame );
	end

	self.faction = GetWatchedFactionInfo();
	
	self.frame:SetPoint( self.layout.Origin.anchor, nUI_Dashboard, "CENTER", self.layout.Origin.xOfs, self.layout.Origin.yOfs );
	
	self.frame:SetBackdrop(
		{
			bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
			edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
			tile     = true, 
			tileSize = 1, 
			edgeSize = 12, 
			insets   = {left = 0, right = 0, top = 0, bottom = 0},
		}
	);

	self.frame:SetBackdropColor( 0, 0, 0, 0.35 );
	self.frame.grid:SetHeight( self.frame:GetHeight() - 2 );
	self.frame.grid:SetWidth( self.frame:GetWidth() - 2 );
	self.frame.grid:SetPoint( "CENTER", self.frame, "CENTER", 0, 0 );
	self.frame.grid:SetFrameLevel( self.frame.bar:GetFrameLevel()+1 );
	self.frame.grid.texture = self.frame.grid:CreateTexture();
	self.frame.grid.texture:SetAllPoints( self.frame.grid );
	self.frame.grid.texture:SetTexture( "Interface\\AddOns\\nUI\\Bars\\Art\\nUI_XPRepOverlay.blp" );
	self.frame.grid.texture:SetTexCoord( 0, 0, 0, 1, 0.978515625, 0, 0.978515625, 1 );
	
	local color = self.layout.Label.Color;
	
	self.frame.bar.center:SetScale( 1.0 / nUI.scale );
	self.frame.bar.center.text:ClearAllPoints();
	self.frame.bar.center.text:SetFont( "Fonts\\ARIALN.TTF", self.layout.Label.size * nUI.scale, "OUTLINE" );
	self.frame.bar.center.text:SetPoint( self.layout.Label.anchor, self.frame, "CENTER", self.layout.Label.xOfs * nUI.scale, self.layout.Label.yOfs * nUI.scale );
	self.frame.bar.center.text:SetTextColor( color.r, color.g, color.b, 1 );
	
	self:updateFaction();	
	self:registerEvents();

end

-------------------------------------------------------------------------------

function prototype:registerEvents()
	
	prototype.super.prototype.registerEvents( self );
	
	self.event_frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.event_frame:RegisterEvent( "UPDATE_FACTION" );
	self.event_frame:RegisterEvent( "CHAT_MSG_COMBAT_FACTION_CHANGE" );
	
end

-------------------------------------------------------------------------------

function prototype:unregisterEvents()
	
	prototype.super.prototype.unregisterEvents( self );
	
	self.event_frame:SetScript( "OnEvent", nil );
	self.event_frame:UnregisterEvent( "UPDATE_FACTION" );
	self.event_frame:UnregisterEvent( "CHAT_MSG_COMBAT_FACTION_CHANGE" );
	
end

-------------------------------------------------------------------------------

function prototype:onEvent()
	
	if event == "UPDATE_FACTION" then
		self:updateFaction();
	elseif event == "CHAT_MSG_COMBAT_FACTION_CHANGE" then
		self:updateFaction( arg1 );
	end	
end

-------------------------------------------------------------------------------

function prototype:updateFaction( msg )

	local	track_faction;
	
	-- if we trapped a combat log chat message, then which faction was it talking about?
	
	if msg then

		local increase_idx = msg:find( " increased" );
		
		if not msg:match( "Reputation with " ) then
			
			return;
			
		elseif not increase_idx then
			
			return;
			
		else
			
			track_faction = msg:sub( 17, increase_idx-1 );
			
			-- if we have a new faction to track, make it the watched faction
			
			if track_faction and self.faction ~= track_faction then
				
				for i=1, GetNumFactions() do
					
					local name = GetFactionInfo( i );
					
					if name == track_faction then
		
						self.faction = track_faction;
						SetWatchedFactionIndex( i );
						
						break;
					end
				end
			end	
		end
	end

	-- who are we actually watching?
	
	local standing, min, max, cur;
	
	self.faction, standing, min, max, cur = GetWatchedFactionInfo();
	
	-- if we still don't have a faction we're watching, then hide the bar
	
	if not self.faction then
		
		if self.frame:IsShown() then self:Hide(); end
		
	else

		if not self.frame:IsShown() then self.frame:Show(); end
		
		self:updateValues( cur, min, min, max );
		self.frame.bar.center.text:SetText( (self.faction..": "..cur.." of "..max.." (%0.1f%% "..(nUI_FactionStandings[standing] or "standing")..")"):format( (cur-min)/(max-min)*100 ) );
		
	end
end