﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------

nUI_Movers = {};

-------------------------------------------------------------------------------

function nUI:initMovers()

	-- create a mover for the capture bars
	
	local frame = WorldStateCaptureBar1;
	local mover = nUI_Movers["nUI_CaptureBarMover"];
	
	nUI_CaptureBarMover = CreateFrame( "Frame", "nUI_CaptureBarMover", UIParent );
	nUI_CaptureBarMover:SetWidth( 173 );
	nUI_CaptureBarMover:SetHeight( 26 );
	
	if not mover then
		nUI_CaptureBarMover:SetPoint( "TOPRIGHT", UIParent, "TOPRIGHT", -10, -20 );
	else
		for i in pairs( mover.Point ) do
			local point = mover.Point[i];
			nUI_CaptureBarMover:SetPoint( point.anchor, point.relative_to, point.relative_pt, point.xOfs, point.yOfs );
		end
	end
	
	nUI_CaptureBarMover:SetScript( "OnEvent", 
		function()	
			if event == "UPDATE_WORLD_STATES" then
				for i=1,NUM_EXTENDED_UI_FRAMES do
					local frame = _G["WorldStateCaptureBar"..i];				
					if not frame.fCache then nUI:lockFrame( frame, true, nil ); end					
					frame.fCache.overlay:SetPoint( "CENTER", nUI_CaptureBarMover, "CENTER", 0, 0 );
				end
			end
		end
	);

	nUI:lockFrame( nUI_CaptureBarMover, true, "Capture Bar" );
	nUI_CaptureBarMover:RegisterEvent( "UPDATE_WORLD_STATES" );	
	
	-- standard Bliz frames we want to  move
	
	QuestWatchFrame:ClearAllPoints();
	QuestWatchFrame:SetPoint( "RIGHT", UIParent, "RIGHT", -5, 256 * nUI.scale );	
	nUI:lockFrame( QuestWatchFrame, true, "Watched Quests" );

	QuestTimerFrame:ClearAllPoints();
	QuestTimerFrame:SetPoint( "TOP", nUI_Console, "BOTTOM", 0, -5 );
	nUI:lockFrame( QuestTimerFrame, true, "Quest Timer" );
	
	DurabilityFrame:ClearAllPoints();
	DurabilityFrame:SetPoint( "TOP", QuestTimerFrame, "BOTTOM", 0, -5 );
	nUI:lockFrame( DurabilityFrame, true, "Durability" );

	WorldStateAlwaysUpFrame:ClearAllPoints();
	WorldStateAlwaysUpFrame:SetPoint( "TOP", nUI_CaptureBarMover, "BOTTOM", 0, -10 );
	nUI:lockFrame( WorldStateAlwaysUpFrame, true, "PvP Objects" );

	GameTooltip:ClearAllPoints();
	GameTooltip:SetPoint( "BOTTOMRIGHT", nUI_Dashboard, "TOPRIGHT", -10, -115 * nUI.scale );
	nUI:lockFrame( GameTooltip, true, "In-Game Tooltips" );
	nUI:lockFrame( GameTooltip, false, "In-Game Tooltips" );

	nUI:manageTooltips();
	
end

-------------------------------------------------------------------------------

function nUI:lockFrame( frame, mode, overlay )

	local new_mover = false;
	
	-- if we don't already have a frame cache for this frame, initialize it
	
	if not frame.fCache then	

		new_mover                   = true;		
		frame.fCache                = {};		
		frame.fCache_SetAllPoints   = frame.SetAllPoints;
		frame.fCache_SetPoint       = frame.SetPoint;
		frame.fCache_ClearAllPoints = frame.ClearAllPoints;

	end
	
	-- lock the frame to prevent anyone else from moving it
	
	if not frame.fCache.locked and mode then
		
		frame.fCache.locked  = true;
		frame.SetAllPoints   = function() end;
		frame.SetPoint       = function() end;
		frame.ClearAllPoints = function() end;
	
	-- make the frame moveable again
	
	elseif frame.fCache.locked and not mode then

		frame.fCache.locked  = false;
		frame.SetAllPoints   = frame.fCache_SetAllPoints;
		frame.SetPoint       = frame.fCache_SetPoint;
		frame.ClearAllPoints = frame.fCache_ClearAllPoints;
		
	end

	-- if this is a new mover, set up the overlay for it
	
	if new_mover then
		
		local frame_name = frame:GetName();
		local mover      = nUI_Movers[frame_name];		

		-- if thise frame has a visible overlay, then add it to the list of 
		-- movers we enable/disable on the "/nui movers" slash command

		if overlay then
			if not nUI.Movers then nUI.Movers = {}; end
			if not nUI.Movers[frame_name] then nUI.Movers[frame_name] = frame; end
		end
		
		-- record the default location for the frame
		
		frame.fCache.DefaultPoints = {};
		
		for i=1,frame:GetNumPoints() do
			
			local point = {};

			point.anchor, point.relative_to, point.relative_pt, point.xOfs, point.yOfs = frame:GetPoint( i );
			point.relative_to = point.relative_to and point.relative_to:GetName() or point_relative_to;
			
			frame.fCache.DefaultPoints[i] = point;
			
		end
		
		-- create a default mover in the config if we don't have one yet
		
		if not mover or not mover.user_placed then
			
			mover = {};
			mover.Point = {};
			mover.user_placed = false;
			
			for i in pairs( frame.fCache.DefaultPoints ) do
				mover.Point[i] = frame.fCache.DefaultPoints[i];
			end
			
			nUI_Movers[frame_name] = mover;
			
		end

		frame.fCache.label = overlay;	
		frame.fCache.drag  = true;
		
		-- create an overlay mover frame to show the user and/or for dragging the frame
	
		frame.fCache.overlay = CreateFrame( "Frame", nil, frame:GetParent() );
		
		frame.fCache.overlay:SetFrameStrata( frame:GetFrameStrata() );
		frame.fCache.overlay:SetFrameLevel( frame:GetFrameLevel()+1 );
		
		frame.fCache.overlay:SetWidth( max( frame:GetWidth(), 25 ) );
		frame.fCache.overlay:SetHeight( max( frame:GetHeight(), 25 ) );

		for i in pairs( mover.Point ) do
			local point = mover.Point[i];
			frame.fCache.overlay:SetPoint( point.anchor, point.relative_to, point.relative_pt, point.xOfs, point.yOfs );
		end
		
		frame.fCache.overlay:EnableMouse( true );
		frame.fCache.overlay:RegisterForDrag();
		frame.fCache.overlay:SetScript( "OnDragStart", function() nUI:dragFrame( frame, true ); end );
		frame.fCache.overlay:SetScript( "OnDragStop", function() nUI:dragFrame( frame, false ); end );
		frame.fCache.overlay:SetMovable( true );

		-- if we want the mover to be visible, creat a texture for it
		
		if overlay then
			
			frame.fCache.overlay:SetBackdrop(
				{
					bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
					edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
					tile     = true, 
					tileSize = 1, 
					edgeSize = 10, 
					insets   = {left = 0, right = 0, top = 0, bottom = 0},
				}
			);
		
			frame.fCache.overlay:SetBackdropColor( 1, 0.75, 0.75, 1 );
			frame.fCache.overlay:SetBackdropBorderColor( 1, 0.75, 0.75, 1 );
			
			frame.fCache.overlay.texture = frame.fCache.overlay:CreateTexture();
			frame.fCache.overlay.texture:SetAllPoints( frame.fCache.overlay );
			frame.fCache.overlay.texture:SetTexture( 1, 0.5, 0.5, 0.5 );
		
		end

		-- lock the frame to the overlay
		
		frame:fCache_ClearAllPoints();
		
		for i in pairs( mover.Point ) do
			local point = mover.Point[i];
			frame:fCache_SetPoint( point.anchor, frame.fCache.overlay, point.anchor, 0, 0 );
		end
		
		-- disable any other movers that may be attached to the frame
		
		frame:RegisterForDrag();
		frame:SetScript( "OnDragStart", nil );
		frame:SetScript( "OnDragStop", nil );

		-- disable dragging
		
		self:enableDrag( frame, false );
	end	
end

-------------------------------------------------------------------------------

function nUI:enableDrag( frame, mode )

	if mode then

		if not frame.fCache.moving and not frame.fCache.drag then
			
			frame.fCache.drag = true;
			frame.fCache.overlay:SetWidth( max( frame:GetWidth(), 25 ) );
			frame.fCache.overlay:SetHeight( max( frame:GetHeight(), 25 ) );
			frame.fCache.overlay:SetScale( frame:GetScale() );
			frame.fCache.overlay:Show();
			frame.fCache.overlay:RegisterForDrag( "LeftButton" );
			frame.fCache.overlay:SetScript( "OnEnter", function() GameTooltip:SetOwner( frame.fCache.overlay, "ANCHOR_BOTTOMRIGHT", 0, 0 ); GameTooltip:SetText( "Left click to drag |cFF00FFFF"..frame.fCache.label.."|r" ); end );
			frame.fCache.overlay:SetScript( "OnLeave", function() GameTooltip:Hide(); end );
			
		end
			
	else
		
		if not frame.fCache.moving and frame.fCache.drag then
			
			frame.fCache.drag = false;
			frame.fCache.overlay:Hide();
			frame.fCache.overlay:RegisterForDrag();
			frame.fCache.overlay:SetScript( "OnEnter", function() end );
			frame.fCache.overlay:SetScript( "OnLeave", function() end );
		end		
	end
end

-------------------------------------------------------------------------------

function nUI:dragFrame( frame, mode )
	
	if mode and frame.fCache.drag and not frame.fCache.moving then
		
		frame.fCache.overlay:StartMoving();
		frame.fCache.moving = true;
		GameTooltip:Hide();
		
	elseif not mode and frame.fCache.drag and frame.fCache.moving then
		
		local mover   = nUI_Movers[frame:GetName()];
		local overlay = frame.fCache.overlay;
		
		overlay:StopMovingOrSizing();
		frame.fCache.moving = false;

		mover.Point = {};
		mover.user_placed = true;
		
		for i=1,overlay:GetNumPoints() do
			
			local point = {};

			point.anchor, point.relative_to, point.relative_pt, point.xOfs, point.yOfs = overlay:GetPoint( i );
			point.relative_to = point.relative_to and point.relative_to:GetName() or point_relative_to;
			
			mover.Point[i] = point;
			
		end
	end
end
 
-------------------------------------------------------------------------------

function nUI:toggleMovers()
	
	if nUI.Movers then
			
		if not nUI.movers_enabled then nUI.movers_enabled = true;
		else nUI.movers_enabled = false;
		end
	
		DEFAULT_CHAT_FRAME:AddMessage( "nUI: Bliz frame movers "..(nUI.movers_enabled and "|cFF00FF00ENABLED|r" or "|cFFFF0000DISABLED|r"), 0.75, 0.75, 0.75 );
		
		for i in pairs( nUI.Movers ) do
			
			nUI:enableDrag( nUI.Movers[i], nUI.movers_enabled );
			
		end
	end
end

-------------------------------------------------------------------------------

function nUI:manageTooltips()
	
	if not nUI_SavedTooltip_SetDefaultAnchor then
			
		nUI_SavedTooltip_SetDefaultAnchor = GameTooltip_SetDefaultAnchor;
		nUI_SavedTooltip_SetOwner         = GameTooltip.SetOwner;
		nUI_SavedTooltip_SmartSetOwner    = SmartSetOwner;
		
		GameTooltip_SetDefaultAnchor = nUI.tooltipSetDefaultAnchor;
		GameTooltip.SetOwner         = nUI.tooltipSetOwner;
		SmartSetOwner                = nUI.tooltipSmartSetOwner;
	
	end
end

-------------------------------------------------------------------------------

function nUI.tooltipSetDefaultAnchor( tooltip, owner )
	
	local mode = nUI_Config.tooltips;

	if mode ~= "default" then
		
		nUI_Tooltip_SetOwner( owner, "ANCHOR_NONE", 0, 0, tooltip );
		
	else
		
		nUI_SavedTooltip_SetDefaultAnchor( tooltip, owner );		

	end	
end

-------------------------------------------------------------------------------

function nUI.tooltipSetOwner( tooltip, owner, anchor, xOfs, yOfs )
	
	local mode = nUI_Config.tooltips;
	
	if mode ~= "default" then		
		
		nUI.tooltipSmartSetOwner( owner, tooltip, xOfs, yOfs );		
		
	else
		
		nUI_SavedTooltip_SetOwner( tooltip, owner, anchor, xOfs, yOfs );
		
	end
	
end

-------------------------------------------------------------------------------

function nUI.tooltipSmartSetOwner( owner, tooltip, xOfs, yOfs )

	if not owner then owner = GameTooltip.fCache.overlay; end
	if not tooltip then tooltip = GameTooltip; end

	if mode ~= "default" then
	
		nUI_Tooltip_SetOwner( owner, "ANCHOR_NONE", xOfs, yOfs, tooltip );
		
	else
		
		nUI_SavedTooltip_SmartSetOwner( owner, "ANCHOR_NONE", xOfs, yOfs, tooltip );

	end
end

-------------------------------------------------------------------------------

function nUI_Tooltip_SetOwner( owner, anchor, xOfs, yOfs, tooltip )

	if not tooltip then tooltip = this; end
	if not xOfs then xOfs = 0; end
	if not yOfs then yOfs = 0; end

	if nUI_Config.tooltips == "fixed" or not owner or owner == UIParent then

		tooltip:Hide();
		
		local overlay = GameTooltip.fCache.overlay;
		
		nUI_SavedTooltip_SetOwner( tooltip, owner or UIParent, "ANCHOR_NONE" );
		
		tooltip:ClearAllPoints();
		tooltip:SetPoint( "BOTTOMRIGHT", UIParent, "BOTTOMLEFT", overlay:GetRight(), overlay:GetBottom() );

	elseif nUI_Config.tooltips == "mouse" then
		
		tooltip:ClearAllPoints();
		
		if tooltip == GameTooltip then
			nUI_SavedTooltip_SetOwner( tooltip, UIParent, "ANCHOR_CURSOR" );
		else
			tooltip:SetParent( UIParent, "ANCHOR_CURSOR" );
		end
		
	else
		
		local x, y         = owner:GetCenter();
		local left         = owner:GetLeft();
		local right        = owner:GetRight();
		local top          = owner:GetTop();
		local bottom       = owner:GetBottom();
		local screenWidth  = UIParent:GetRight() - UIParent:GetLeft();
		local screenHeight = UIParent:GetTop() - UIParent:GetBottom();
		local scale        = owner:GetEffectiveScale();
		local point        = "";
		
		if x ~= nil and y ~= nil and left ~= nil and right ~= nil and top ~= nil 
		and bottom ~= nil and screenWidth > 0 and screenHeight > 0 then
			
			xOfs   = xOfs * scale;
			yOfs   = yOfs * scale;
			x      = x * scale;
			y      = y * scale;		
			left   = left * scale;
			right  = right * scale;
			width  = right - left;
			top    = top * scale;
			bottom = bottom * scale;		
			
			if y <= (screenHeight * (1/2)) then
				
				top   = top + yOfs;
				point = "TOP";
				
				if (top < 0) then yOfs = yOfs - top; end
				
			else
				
				yOfs   = -yOfs;
				bottom = bottom + yOfs;
				point  = "BOTTOM";
				
				if bottom > screenHeight then yOfs = yOfs + (screenHeight - bottom); end
			end
			
			if x <= (screenWidth * (1/2)) then
				
				left = left + xOfs;
				
				if point == "BOTTOM" then
					
					point = point.."RIGHT";
					xOfs  = xOfs - width;
					
					if left < 0 then xOfs = xOfs - left; end
					
				else
					
					point = point.."LEFT";
					
					if left < 0 then xOfs = xOfs - left; end
				end
				
			else
				
				xOfs  = -xOfs;
				right = right + xOfs;
				
				if point == "BOTTOM" then
					
					point = point.."LEFT";
					xOfs  = xOfs + width;
					
					if right > screenWidth then xOfs = xOfs - (right - screenWidth); end
					
				else
	
					point = point.."RIGHT";
					
					if right > screenWidth then xOfs = xOfs + (screenWidth - right); end
				end
			end
			
			if point == "" then point = "TOPLEFT"; end
			
			scale = tooltip:GetScale();
			
			if scale then
				xOfs = xOfs / scale;
				yOfs = yOfs / scale;
			end
		
			if tooltip == GameTooltip then
				nUI_SavedTooltip_SetOwner( tooltip, owner, "ANCHOR_"..point, xOfs, yOfs );
			else
				tooltip:SetOwner( owner, "ANCHOR_"..point, xOfs, yOfs );
			end
		end
	end
end